/*
############################################################################
# This file contains the word functions for the pfth Forth interpreter.
#
# Copyright (c) 2012 Dave Hein
# MIT Licensed
############################################################################
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include "pfth.h"
#include "words.h"

char pad[256];
char *printptr;

int herefunc(void *parm)
{
    push((int)here);
    return errorcode;
}

int dotfunc(void *parm)
{
    int len = ConvertSignedNumber(pop());
    Type(printptr+1, len);
    printf(" ");
    return errorcode;
}

int immediatefunc(void *parm)
{
    if (lastword) lastword->flags |= FLAG_IMMEDIATE;
    return errorcode;
}

int allotfunc(void *parm)
{
    here += pop();
    return errorcode;
}

int compileonfunc(void *parm)
{
    compileflag = -1;
    return errorcode;
}

int compileofffunc(void *parm)
{
    compileflag = 0;
    return errorcode;
}

int FAST rotfunc(void *parm)
{
    int value3 = pop();
    int value2 = pop();
    int value1 = pop();
    push(value2);
    push(value3);
    push(value1);
    return errorcode;
}

int FAST dupfunc(void *parm)
{
    int value = pop();
    push(value);
    push(value);
    return errorcode;
}

int FAST swapfunc(void *parm)
{
    int value1 = pop();
    int value2 = pop();
    push(value1);
    push(value2);
    return errorcode;
}

int twoslashfunc(void *parm)
{
    int value = pop();
    push(value >> 1);
    return errorcode;
}

int postponefunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    dictentryT *wordentry = FindWord(wordptr, len);
    if (!wordentry)
    {
        Type(wordptr, len);
        printf(" ?\n");
        return RETURN_ERROR;
    }
    *((int *)here) = (int)wordentry;
    here += 4;
    return errorcode;
}

int beginfunc(void *parm)
{
    if (!compileflag || !evaluate)
        return RETURN_ERROR;
    retstk[retidx++] = (int)(here - 4);
    return errorcode;
}

int untilfunc(void *parm)
{
    *(int *)here = (int)_jzword;
    here += 4;
    *(int *)here = (int)retstk[--retidx];
    here += 4;
    return errorcode;
}

int againfunc(void *parm)
{
    *(int *)here = (int)_jmpword;
    here += 4;
    *(int *)here = (int)retstk[--retidx];
    here += 4;
    return errorcode;
}

int FAST dofunc(void *parm)
{
    if (compileflag && evaluate)
    {
        *(int *)here = (int)parm;
        retstk[retidx++] = (int)(here + 1);
        here += 4;
    }
    else
    {
        int index = pop();
        int limit = pop();
        pushret(limit);
        pushret(index);
    }
    return errorcode;
}

int FAST whilefunc(void *parm)
{
    *(int *)here = (int)_jzword;
    here += 4;
    retstk[retidx++] = (int)here;
    here += 4;
    return errorcode;
}

int recursefunc(void *parm)
{
    *(int *)here = (int)lastword;
    here += 4;
    return errorcode;
}

int FAST repeatfunc(void *parm)
{
    int *ptr = (int *)retstk[--retidx];
    *(int *)here = (int)_jmpword;
    here += 4;
    *(int *)here = (int)retstk[--retidx];
    *ptr = (int)here;
    here += 4;
    return errorcode;
}

int iffunc(void *parm)
{
    here = ALIGN(here);
    *((int *)here) = (int)_jzword;
    here += 4;
    retstk[retidx++] = (int)here;
    *((int *)here) = 12345678;
    here += 4;
    return errorcode;
}

int elsefunc(void *parm)
{
    int *ptr = (int *)retstk[--retidx];
    *ptr = (int)(here + 4);
    here = ALIGN(here);
    *((int *)here) = (int)_jmpword;
    here += 4;
    retstk[retidx++] = (int)here;
    *((int *)here) = 12345678;
    here += 4;
    return errorcode;
}

int thenfunc(void *parm)
{
    int *ptr = (int *)retstk[--retidx];
    *ptr = (int)(here - 4);
    return errorcode;
}

int literalfunc(void *parm)
{
    here = ALIGN(here);
    *((int *)here) = (int)_litword;
    here += 4;
    *((int *)here) = pop();
    here += 4;
    return errorcode;
}

int doesfunc(void *parm)
{
    if (lastword)
    {
        lastword->does = (int *)(execaddr + 1);
        return RETURN_EXIT;
    }
    return RETURN_ERROR;
}

int udotfunc(void *parm)
{
    int len = ConvertUnsignedNumber(pop());
    Type(printptr+1, len);
    printf(" ");
    return errorcode;
}

int ulessfunc(void *parm)
{
    unsigned int value2 = pop();
    unsigned int value1 = pop();
    push(-(value1 < value2));
    return errorcode;
}

int typefunc(void *parm)
{
    int num = pop();
    char *ptr = (char *)pop();
    Type(ptr, num);
    return errorcode;
}

int sourcefunc(void *parm)
{
    push((int)inputptr);
    push(inputlen);
    return errorcode;
}

int _jmpfunc(void *parm)
{
    execaddr = (int *)execaddr[1];
    return errorcode;
}

int _jzfunc(void *parm)
{
    if (stkidx < 1) return RETURN_ERROR;
    if (stack[--stkidx])
        execaddr++;
    else
        execaddr = (int *)execaddr[1];
    return errorcode;
}

int tickfunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    push((int)FindWord(wordptr, len));
    return errorcode;
}

int parenfunc(void *parm)
{
    int len;
    char *wordptr;

    while (1)
    {
        if (inputidx >= inputlen)
        {
            if (!Refill()) return RETURN_ERROR;
            if (inputlen && inputptr[0] == ')')
            {
                inputidx = 1;
                break;
            }
        }
        wordptr = WordDelimiter(&len, ')');
        if (wordptr[len] == ')') break;
    }

    return errorcode;
}

int bodyfunc(void *parm)
{
    dictentryT *entry = (dictentryT *)pop();
    push((int)ALIGN(entry->name + entry->len));
    return errorcode;
}

int charfunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    push((int)*wordptr);
    return errorcode;
}

int dotquotefunc(void *parm)
{
    if (compileflag && evaluate)
    {
        CompileString(parm);
    }
    else
    {
        char *ptr = (char *)(++execaddr);
        int num = *ptr++;
        Type(ptr, num);
        execaddr += num/4;
    }
    return errorcode;
}

int abortquotefunc(void *parm)
{
    if (compileflag && evaluate)
    {
        CompileString(parm);
    }
    else
    {
        char *ptr = (char *)(++execaddr);
        int num = *ptr++;
        if (pop())
        {
            Type(ptr, num);
            stkidx = 0;
            retidx = 0;
            return RETURN_ERROR;
        }
        execaddr += num/4;
    }
    return errorcode;
}

int _litquotefunc(void *parm)
{
    char *ptr = (char *)(++execaddr);
    int num = *ptr++;
    push((int)ptr);
    push(num);
    execaddr += num/4;
    return errorcode;
}

int squotefunc(void *parm)
{
    CompileString(_litquoteword);
    return errorcode;
}

int lessnumberfunc(void *parm)
{
    printptr = pad + 255;
    return errorcode;
}

int numberfunc(void *parm)
{
    int64_t value = pop2();
    push2(value/base);
    *printptr-- = digit2ascii((int)(value % base));
    return errorcode;
}

int numbersfunc(void *parm)
{
    uint64_t value = (uint64_t)pop2();

    *printptr-- = digit2ascii((int)(value % base));
    while ((value /= base))
    {
        *printptr-- = digit2ascii((int)(value % base));
    }
    push2(0);
    return errorcode;
}

int numbergreaterfunc(void *parm)
{
    pop2();
    push((int)(printptr + 1));
    push((int)(pad + 255 - printptr));
    return errorcode;
}

int holdfunc(void *parm)
{
    *printptr-- = pop();
    return errorcode;
}

int signfunc(void *parm)
{
    if (pop())
        *printptr-- = '-';
    return errorcode;
}

int starslashfunc(void *parm)
{
    int64_t value3 = (int64_t)pop();
    int64_t value2 = (int64_t)pop();
    int64_t value1 = (int64_t)pop();
    push((int)((value1 * value2) / value3));
    return errorcode;
}

int starslashmodfunc(void *parm)
{
    int64_t value3 = (int64_t)pop();
    int64_t value2 = (int64_t)pop();
    int64_t value1 = (int64_t)pop();
    value1 *= value2;
    push((int)(value1 % value3));
    push((int)(value1 / value3));
    return errorcode;
}

int greaternumberfunc(void *parm)
{
    int digit;
    int num = pop();
    char *ptr = (char *)pop();
    int64_t value = pop2();
    while (num > 0)
    {
        if (*ptr >= '0' && *ptr <= '9')
            digit = *ptr - '0';
        else if (*ptr >= 'A' && *ptr <= 'Z')
            digit = *ptr - 'A' + 10;
        else if (*ptr >= 'a' && *ptr <= 'z')
            digit = *ptr - 'a' + 10;
        else
            break;
        if (digit >= base) break;
        value = (value * base) + digit;
        ptr++;
        num--;
    }
    push2(value);
    push((int)ptr);
    push(num);
    return errorcode;
}

int fmslashmodfunc(void *parm)
{
    int64_t value2 = (int64_t)pop();
    int64_t value1 = pop2();
    int64_t quotient, remainder, round;
    int sign2 = (value2 < 0);
    int sign1 = (value1 < 0);

    if (sign1 == sign2)
    {
        quotient = value1 / value2;
        remainder = value1 % value2;
    }
    else
    {
        round = (value1 << 1) - 1;
        quotient = (value1 - value2 + round) / value2;
        remainder = value1 - (quotient * value2);
    }
    push((int)remainder);
    push((int)quotient);
    return errorcode;
}

int smslashremfunc(void *parm)
{
    int64_t value2 = (int64_t)pop();
    int64_t value1 = pop2();
    push((int)(value1 % value2));
    push((int)(value1 / value2));
    return errorcode;
}

int unsignedmslashmodfunc(void *parm)
{
    uint64_t value2 = (uint64_t)(unsigned int)pop();
    uint64_t value1 = (uint64_t)pop2();
    push((int)(value1 % value2));
    push((int)(value1 / value2));
    return errorcode;
}

int bracketcharfunc(void *parm)
{
    int len;
    char *wordptr;

    if (!compileflag || !evaluate)
        return RETURN_ERROR;

    wordptr = Word(&len);

    if (!len)
        return RETURN_ERROR;

    here = ALIGN(here);
    *((int *)here) = (int)_litword;
    here += 4;
    *((int *)here) = *wordptr;
    here += 4;

    return errorcode;
}

int brackettickfunc(void *parm)
{
    int len;
    char *wordptr;
    dictentryT *wordentry = 0;

    if (!compileflag || !evaluate)
        return RETURN_ERROR;

    wordptr = Word(&len);
    wordentry = FindWord(wordptr, len);

    if (!wordentry)
        return RETURN_ERROR;

    here = ALIGN(here);
    *((int *)here) = (int)_litword;
    here += 4;
    *((int *)here) = (int)wordentry;
    here += 4;

    return errorcode;
}

int decimalfunc(void *parm)
{
    base = 10;
    return errorcode;
}

int FAST ifunc(void *parm)
{
    stack[stkidx++] = retstk[retidx-1];
    return errorcode;
}

int FAST jfunc(void *parm)
{
    stack[stkidx++] = retstk[retidx-3];
    return errorcode;
}

int twodupfunc(void *parm)
{
    int value1 = pop();
    int value2 = pop();
    push(value2);
    push(value1);
    push(value2);
    push(value1);
    return errorcode;
}

int overfunc(void *parm)
{
    int value1 = pop();
    int value2 = pop();
    push(value2);
    push(value1);
    push(value2);
    return errorcode;
}

int movefunc(void *parm)
{
    int num = pop();
    int addr2 = pop();
    int addr1 = pop();
    memmove((char *)addr2, (char *)addr1, num);
    return errorcode;
}

int leavefunc(void *parm)
{
    if (compileflag && evaluate)
    {
        int index;
        *(int *)here = (int)parm;
        here += 4;
        for (index = retidx++; index > 0; index--)
        {
            if (((int)retstk[index-1]) & 3)
            {
                retstk[index] = (int)(here + 3);
                break;
            }
            retstk[index] = retstk[index-1];
        }
        here += 4;
    }
    else
    {
        popret();
        popret();
        execaddr = (int *)execaddr[1];
    }
    return errorcode;
}

int FAST loopfunc(void *parm)
{
    if (compileflag && evaluate)
    {
        int flag;
        int *ptr;
        *(int *)here = (int)parm;
        here += 4;
        while (1)
        {
            flag = (int)retstk[--retidx];
            ptr = (int *)(flag & ~3);
            flag &= 3;
            if (flag > 1) *ptr = (int)here;
            if (flag != 3) break;
        }
        *(int *)here = (int)ptr;
        here += 4;
    }
    else
    {
        int index, limit;
        
        if (retidx < 2)
            return RETURN_ERROR;

        index = retstk[retidx-1];
        limit = retstk[retidx-2];
        
        if (++index < limit)
        {
            execaddr = (int *)execaddr[1];
            retstk[retidx-1] = index;
        }
        else
        {
            execaddr++;
            retidx -= 2; 
        }       
    }
    return errorcode;
}

int FAST plusloopfunc(void *parm)
{
    if (compileflag && evaluate)
    {
        int flag;
        int *ptr;
        *(int *)here = (int)parm;
        here += 4;
        while (1)
        {
            flag = (int)retstk[--retidx];
            ptr = (int *)(flag & ~3);
            flag &= 3;
            if (flag > 1) *ptr = (int)here;
            if (flag != 3) break;
        }
        *(int *)here = (int)ptr;
        here += 4;
    }
    else
    {
        int index = popret();
        int limit = popret();

        index += pop();

        if (index < limit)
        {
            execaddr = (int *)execaddr[1];
            pushret(limit);
            pushret(index);
        }
        else
            execaddr++;
    }
    return errorcode;
}

void InitializeCoreWords()
{
    MakeEntry("(", parenfunc, FLAG_IMMEDIATE);
    MakeEntry("_jmp", _jmpfunc, FLAG_LITERAL | FLAG_JUMP);
    _jmpword = lastword;
    MakeEntry("_jz", _jzfunc, FLAG_LITERAL | FLAG_JUMP);
    _jzword = lastword;
    MakeEntry("literal", literalfunc, FLAG_IMMEDIATE);
    MakeEntry("immediate", immediatefunc, 0);
    MakeEntry("postpone", postponefunc, FLAG_IMMEDIATE);
    MakeEntry("does>", doesfunc, 0);
    MakeEntry("_lit\"", _litquotefunc, FLAG_STRING);
    _litquoteword = lastword;
    MakeEntry("u<", ulessfunc, 0);
    MakeEntry(">body", bodyfunc, 0);
    MakeEntry("do", dofunc, FLAG_IMMEDIATE);
    MakeEntry("leave", leavefunc, FLAG_IMMEDIATE | FLAG_LITERAL | FLAG_JUMP);
    MakeEntry("loop", loopfunc, FLAG_IMMEDIATE | FLAG_LITERAL | FLAG_JUMP);
    MakeEntry("+loop", plusloopfunc, FLAG_IMMEDIATE | FLAG_LITERAL | FLAG_JUMP);
    MakeEntry("abort\"", abortquotefunc, FLAG_IMMEDIATE | FLAG_STRING);
    MakeEntry("move", movefunc, 0);
    MakeEntry(".\"", dotquotefunc, FLAG_IMMEDIATE | FLAG_STRING);
    MakeEntry("'", tickfunc, 0);
    MakeEntry("if", iffunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("else", elsefunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("then", thenfunc, FLAG_IMMEDIATE);
    MakeEntry("begin", beginfunc, FLAG_IMMEDIATE);
    MakeEntry("while", whilefunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("repeat", repeatfunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("until", untilfunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("again", againfunc, FLAG_IMMEDIATE | FLAG_LITERAL);
    MakeEntry("u.", udotfunc, 0);
    MakeEntry("2/", twoslashfunc, 0);
    MakeEntry("allot", allotfunc, 0);
    MakeEntry("here", herefunc, 0);
    MakeEntry("source", sourcefunc, 0);
    MakeEntry("sm/rem", smslashremfunc, 0);
    MakeEntry("char", charfunc, 0);
    MakeEntry("s\"", squotefunc, FLAG_IMMEDIATE | FLAG_STRING);
    MakeEntry("recurse", recursefunc, FLAG_IMMEDIATE);
    MakeEntry("<#", lessnumberfunc, 0);
    MakeEntry("#", numberfunc, 0);
    MakeEntry("#>", numbergreaterfunc, 0);
    MakeEntry("#s", numbersfunc, 0);
    MakeEntry("hold", holdfunc, 0);
    MakeEntry("sign", signfunc, 0);
    MakeEntry("*/", starslashfunc, 0);
    MakeEntry("*/mod", starslashmodfunc, 0);
    MakeEntry(">number", greaternumberfunc, 0);
    MakeEntry("fm/mod", fmslashmodfunc, 0);
    MakeEntry("[char]", bracketcharfunc, FLAG_IMMEDIATE);
    MakeEntry("[']", brackettickfunc, FLAG_IMMEDIATE);
    MakeEntry("decimal", decimalfunc, 0);
    MakeEntry("i", ifunc, 0);
    MakeEntry("j", jfunc, 0);
    MakeEntry("rot", rotfunc, 0);
    MakeEntry("dup", dupfunc, 0);
    MakeEntry("swap", swapfunc, 0);
    MakeEntry("over", overfunc, 0);
    MakeEntry("2dup", twodupfunc, 0);
    MakeEntry("type", typefunc, 0);
    MakeEntry("[", compileofffunc, FLAG_IMMEDIATE);
    MakeEntry("]", compileonfunc, FLAG_IMMEDIATE);
    MakeEntry(".", dotfunc, 0);
    MakeEntry("um/mod", unsignedmslashmodfunc, 0);
}
/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
