/*
############################################################################
# This file contains the word functions for the pfth Forth interpreter.
#
# Copyright (c) 2012 Dave Hein
# MIT Licensed
############################################################################
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include "pfth.h"
#include "words.h"

int dslashfunc(void *parm)
{
    int64_t value2 = pop2();
    int64_t value1 = pop2();
    push2(value1 / value2);
    return errorcode;
}

int udslashfunc(void *parm)
{
    uint64_t value2 = pop2();
    uint64_t value1 = pop2();
    push2(value1 / value2);
    return errorcode;
}

int dstarfunc(void *parm)
{
    int64_t value2 = pop2();
    int64_t value1 = pop2();
    push2(value1 * value2);
    return errorcode;
}

int dplusfunc(void *parm)
{
    int64_t value2 = pop2();
    int64_t value1 = pop2();
    push2(value1 + value2);
    return errorcode;
}

int dmodfunc(void *parm)
{
    int64_t value2 = pop2();
    int64_t value1 = pop2();
    push2(value1 % value2);
    return errorcode;
}

int udmodfunc(void *parm)
{
    uint64_t value2 = pop2();
    uint64_t value1 = pop2();
    push2(value1 % value2);
    return errorcode;
}

int mstarfunc(void *parm)
{
    int64_t value2 = (int64_t)pop();
    int64_t value1 = (int64_t)pop();
    push2(value1 * value2);
    return errorcode;
}

int unsignedmstarfunc(void *parm)
{
    uint64_t value2 = (uint64_t)(unsigned int)pop();
    uint64_t value1 = (uint64_t)(unsigned int)pop();
    push2(value1 * value2);
    return errorcode;
}

void InitializeDoubleWords()
{
    MakeEntry("m*", mstarfunc, 0);
    MakeEntry("um*", unsignedmstarfunc, 0);
    MakeEntry("d+", dplusfunc, 0);
    MakeEntry("d*", dstarfunc, 0);
    MakeEntry("d/", dslashfunc, 0);
    MakeEntry("dmod", dmodfunc, 0);
    MakeEntry("ud/", udslashfunc, 0);
    MakeEntry("udmod", udmodfunc, 0);
}
/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
