/*
############################################################################
# This file contains the word functions for the pfth Forth interpreter.
#
# Copyright (c) 2012 Dave Hein
# MIT Licensed
############################################################################
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include "pfth.h"
#include "words.h"

char pad[256];
char *printptr;

int dotrfunc(void *parm)
{
    int value2 = pop();
    int value1 = pop();
    int len = ConvertSignedNumber(value1);

    while (len < value2--) printf(" ");
    Type(printptr+1, len);
    printf(" ");
    return errorcode;
}

int FAST questiondofunc(void *parm)
{
    if (compileflag && evaluate)
    {
        *(int *)here = (int)parm;
        here += 4;
        retstk[retidx++] = (int)(here + 2);
        here += 4;
    }
    else
    {
        int index = pop();
        int limit = pop();
        if (index < limit)
        {
            pushret(limit);
            pushret(index);
            execaddr++;
        }
        else
        {
            execaddr = (int *)execaddr[1];
        }
    }
    return errorcode;
}

int backslashfunc(void *parm)
{
    int len;

    WordDelimiter(&len, 256);

    return errorcode;
}

int mstarslashfunc(void *parm)
{
    int64_t value3 = (int64_t)pop();
    int64_t value2 = (int64_t)pop();
    int64_t value1 = pop2();
    push2((value1 * value2) / value3);
    return errorcode;
}

int FAST tofunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    dictentryT *wordentry = FindWord(wordptr, len);

    if (!wordentry)
    {
        Type(wordptr, len);
        printf(" ?\n");
        return RETURN_ERROR;
    }

    if (compileflag && evaluate)
    {
        here = ALIGN(here);
        *((int *)here) = (int)_litword;
        here += 4;
        *((int *)here) = (int)BODY(wordentry);
        here += 4;
        *((int *)here) = (int)storeword;
        here += 4;
    }
    else
    {
        int *ptr = (int *)BODY(wordentry);
        if (stkidx < 1)
            return RETURN_ERROR;
        *ptr = stack[--stkidx];
    }

    return errorcode;
}

int dumpfunc(void *parm)
{
    int i;
    int num = pop();
    unsigned char *ptr = (unsigned char *)pop();

    if (errorcode == RETURN_ERROR) return errorcode;

    for (i = 0; i < num; i++)
    {
        if (!(i & 15)) printf("\n%8.8x  ", (int)ptr);
        printf(" %2.2x", *ptr++);
    }

    printf("\n\n");

    return errorcode;
}

int wordsfunc(void *parm)
{
    int len = 0;
    dictentryT *wordentry = lastword;

    while (wordentry)
    {
        if (len + wordentry->len > 78)
        {
            printf("\n");
            len = wordentry->len;
        }
        else if (len)
        {
            printf(" ");
            len += wordentry->len + 1;
        }
        else
            len = wordentry->len;
        Type(wordentry->name, wordentry->len);
        wordentry = wordentry->prev;
    }
    printf("\n");
    return errorcode;
}

int forgetfunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    dictentryT *wordentry = FindWord(wordptr, len);

    if (!wordentry)
    {
        Type(wordptr, len);
        printf(" ?\n");
        return RETURN_ERROR;
    }

    here = (char *)wordentry;
    lastword = wordentry->prev;

    return errorcode;
}

int _markerfunc(void *parm)
{
    dictentryT *wordentry = (dictentryT *)parm;
    here = (char *)wordentry;
    lastword = wordentry->prev;
    return errorcode;
}

int markerfunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    Create(wordptr, len, _markerfunc);
    return errorcode;
}

int dotsfunc(void *parm)
{
    int i, len;
    printf("<");
    len = ConvertSignedNumber(stkidx);
    Type(printptr+1, len);
    printf("> ");
    for (i = 0; i < stkidx; i++)
    {
        len = ConvertSignedNumber(stack[i]);
        Type(printptr+1, len);
        printf(" ");
    }
    printf("\n");
    return errorcode;
}

#if 0
int unusedfunc(void *parm)
{
    push((int)(((char *)&codespace[CODE_SIZE]) - here));
    return errorcode;
}
#endif

int hexfunc(void *parm)
{
    base = 16;
    return errorcode;
}

int withinfunc(void *parm)
{
    int value3 = pop();
    int value2 = pop();
    int value1 = pop();
    if (value2 < value3)
        push(-(value2 <= value1 && value1 < value3));
    else if (value2 > value3)
        push(-(value2 <= value1 || value1 < value3));
    else
        push(0);
    return errorcode;
}

void InitializeExtendedWords()
{
    MakeEntry("forget", forgetfunc, 0);
    MakeEntry("\\", backslashfunc, FLAG_IMMEDIATE);
    MakeEntry("dump", dumpfunc, 0);
    MakeEntry("words", wordsfunc, 0);
    MakeEntry("?do", questiondofunc, FLAG_IMMEDIATE | FLAG_LITERAL | FLAG_JUMP);
    MakeEntry("m*/", mstarslashfunc, 0);
    //MakeEntry("unused", unusedfunc, 0);
    MakeEntry("within", withinfunc, 0);
    MakeEntry("hex", hexfunc, 0);
    MakeEntry("marker", markerfunc, 0);
    MakeEntry(".r", dotrfunc, 0);
    MakeEntry("to", tofunc, FLAG_IMMEDIATE);
    MakeEntry(".s", dotsfunc, 0);
}
/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
