{
############################################################################
# PFTH - This program implements a Forth interpreter.
#
# Copyright (c) 2012 Dave Hein
# MIT Licensed
############################################################################
}
con
  _clkmode = xtal1+pll16x
  _clkfreq = 80_000_000

  FLAG_IMMEDIATE = 1

'*******************************************************************************
' This Spin code waits three seconds and then starts the Forth cog
'*******************************************************************************
pub main
  waitcnt(clkfreq*3+cnt)
  cognew(@forth, 0)

'*******************************************************************************
' pfth cog code
'*******************************************************************************
dat
                        org     0                

forth
parm                    mov     outa, outbit
temp                    mov     dira, outbit
                        jmp     #innerloop              ' Begin executing the boot interpreter

'*******************************************************************************
' Execute the words contained in the body of a word
' Changes parm, temp1
'*******************************************************************************
innerloopfunc           call    #body1
innerloopcall           wrlong  pc, returnptr
                        add     returnptr, #4
                        mov     pc, parm1
                        
innerloop               rdlong  parm, pc                wz
 if_z                   jmp     #exitfunc
                        add     parm, #4
                        rdlong  temp1, parm
                        add     pc, #4
                        jmp     temp1
                        
pc                      long    @_xboot1+16             ' Program Counter

'*******************************************************************************
' Stop executing the current word, and return to the calling word
' No Changes              
'*******************************************************************************
exitfunc                sub     returnptr, #4
                        rdlong  pc, returnptr
                        jmp     #innerloop

'*******************************************************************************
' Abort or quit execution, and return to the interpreter
' No Changes
'*******************************************************************************
abortfunc               mov     stackptr, stackptr0
quitfunc                mov     returnptr, returnptr0
                        add     returnptr, #4           ' Use second entry return stack
                        rdlong  pc, returnptr      
                        jmp     #innerloop

'*******************************************************************************
' Push the address of the word's body onto the stack
' Execute the words pointed to by the does pointer, if non-zero
' No changes
'*******************************************************************************
varfunc                 mov     parm2, parm
                        add     parm2, #4               ' DOES> pointer
                        call    #body1
                        call    #push
                        rdlong  parm1, parm2            wz
        if_z            jmp     #innerloop              ' Done with varfunc
                        jmp     #innerloopcall          ' Execute DOES> code

'*******************************************************************************
' Execute the word on the stack
' Changes parm, temp1
'*******************************************************************************
executefunc             sub     stackptr, #4
                        rdlong  parm, stackptr
                        add     parm, #4
                        rdlong  temp1, parm
                        jmp     temp1                   ' Execute code

'*******************************************************************************
' Get the next word from the input buffer using the delimiter from the stack
' Changes parm, parm1, parm2, temp1, temp2
'*******************************************************************************
wordfunc                sub     stackptr, #4
                        rdlong  parm, stackptr
                        call    #word_del
                        mov     temp1, #1
                        shl     temp1, #15
                        sub     temp1, parm2
                        sub     temp1, #1
                        wrlong  temp1, stackptr
                        add     stackptr, #4
                        wrbyte  parm2, temp1
                        cmps    parm2, #0               wc, wz
        if_c_or_z       jmp     #innerloop
:loop                   add     temp1, #1
                        rdbyte  temp2, parm1
                        add     parm1, #1
                        wrbyte  temp2, temp1
                        djnz    parm2, #:loop
                        jmp     #innerloop

'*******************************************************************************
' Find the word specfied on the stack in the dictionary
' Changes parm1, parm2, temp4
'*******************************************************************************
findfunc                call    #pop
                        mov     temp4, parm1
                        add     parm1, #1
                        rdbyte  parm2, temp4
                        call    #findword
                        mov     parm1, parm             wz
        if_z            jmp     #findfunc1
                        call    #push
                        add     parm1, #12
                        rdbyte  parm1, parm1
                        and     parm1, #1
                        shl     parm1, #1
                        sub     parm1, #1
                        call    #push
                        jmp     #innerloop
findfunc1               mov     parm1, temp4
                        call    #push
                        mov     parm1, #0
                        call    #push
                        jmp     #innerloop

'*******************************************************************************
' Refill the input buffer
' Changes none
'*******************************************************************************
refillfunc              call    #refill
                        jmp     #innerloop

'*******************************************************************************
' Send the character from the stack to the output port
' Changes parm
'*******************************************************************************
emitfunc                call    #pop
                        mov     parm, parm1
                        call    #putch
                        jmp     #innerloop

'*******************************************************************************
' Get a character from the input port and put it on the stack
' Changes parm, parm1
'*******************************************************************************
keyfunc                 call    #getch
                        mov     parm1, parm
                        call    #push
                        jmp     #innerloop

'*******************************************************************************
' Get a string from the input port
' Changes parm, parm1, parm2
'*******************************************************************************
acceptfunc              call    #pop2
                        call    #getstr
                        call    #push
                        jmp     #innerloop

'*******************************************************************************
' Get an address and value from the stack, and store the value at the address
' No changes              
'*******************************************************************************
storefunc               call    #pop2
                        wrlong  parm1, parm2
                        jmp     #innerloop

'*******************************************************************************
' Fetch a value from the address specified on the stack, and put it on the stack
' Changes parm1              
'*******************************************************************************
fetchfunc               call    #pop
                        rdlong  parm1, parm1
                        call    #push
                        jmp     #innerloop

'*******************************************************************************
' Get an address and byte from the stack, and store the byte at the address
' No changes              
'*******************************************************************************
cstorefunc              call    #pop2
                        wrbyte  parm1, parm2
                        jmp     #innerloop

'*******************************************************************************
' Fetch a byre from the address specified on the stack, and put it on the stack
' Changes parm1              
'*******************************************************************************
cfetchfunc              call    #pop
                        rdbyte  parm1, parm1
                        call    #push
                        jmp     #innerloop
              
'*******************************************************************************
' Add two values from the stack, and write the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
plusfunc                call    #pop2
                        add     parm1, parm2
                        call    #push
                        jmp     #innerloop
              
'*******************************************************************************
' Subtract two values from the stack, and write the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
minusfunc               call    #pop2
                        sub     parm1, parm2
                        call    #push
                        jmp     #innerloop
              
'*******************************************************************************
' Multiply two values from the stack, and write the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
multfunc                call    #pop2
                        call    #multiply
                        call    #push
                        jmp     #innerloop
              
'*******************************************************************************
' Divide two values from the stack, and write the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
dividefunc              call    #pop2
                        call    #divide
                        mov     parm1, parm2
                        call    #push
                        jmp     #innerloop
              
'*******************************************************************************
' Compute the modulus from two values from the stack, and write the result back
' to the stack
' Changes parm1, parm2              
'*******************************************************************************
modfunc                 call    #pop2
                        call    #divide
                        call    #push
                        jmp     #innerloop
              
'*******************************************************************************
' Compare two values from the stack to determine if the second one is less than
' the first one, and write the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
lessfunc                call    #pop2
                        cmps    parm1, parm2            wc
        if_c            neg     parm1, #1
        if_nc           mov     parm1, #0
                        call    #push
                        jmp     #innerloop
              
'*******************************************************************************
' Compare two values from the stack to determine if they are equal, and write
' the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
equalfunc               call    #pop2
                        cmp     parm1, parm2            wz
        if_z            neg     parm1, #1
        if_nz           mov     parm1, #0
                        call    #push
                        jmp     #innerloop
              
'*******************************************************************************
' Compare two values from the stack to determine if the second one is greater
' than the first one, and write the result back to the stack
' Changes parm1, parm2              
'*******************************************************************************
greaterfunc             call    #pop2
                        cmps    parm1, parm2            wc, wz
        if_nz_and_nc    neg     parm1, #1
        if_z_or_c       mov     parm1, #0
                        call    #push
                        jmp     #innerloop

'*******************************************************************************
' Compute the logical AND of two values from the stack, and write the result
' back to the stack
' Changes parm1, parm2              
'*******************************************************************************
andfunc                 call    #pop2
                        and     parm1, parm2
                        call    #push
                        jmp     #innerloop
              
'*******************************************************************************
' Compute the logical OR of two values from the stack, and write the result
' back to the stack
' Changes parm1, parm2              
'*******************************************************************************
orfunc                  call    #pop2
                        or      parm1, parm2
                        call    #push
                        jmp     #innerloop
              
'*******************************************************************************
' Compute the logical XOR of two values from the stack, and write the result
' back to the stack
' Changes parm1, parm2              
'*******************************************************************************
xorfunc                 call    #pop2
                        xor     parm1, parm2
                        call    #push
                        jmp     #innerloop
              
'*******************************************************************************
' Right-shift the second value on the stack by the number of bits specified by
' the first value on the stack, and write the result to the stack
' Changes parm1, parm2              
'*******************************************************************************
rshiftfunc              call    #pop2
                        shr     parm1, parm2
                        call    #push
                        jmp     #innerloop
              
'*******************************************************************************
' Left-shift the second value on the stack by the number of bits specified by
' the first value on the stack, and write the result to the stack
' Changes parm1, parm2              
'*******************************************************************************
lshiftfunc              call    #pop2
                        shl     parm1, parm2
                        call    #push
                        jmp     #innerloop

'*******************************************************************************
' Push the stack depth to the stack
' Changes parm1              
'*******************************************************************************
depthfunc               mov     parm1, stackptr
                        sub     parm1, stackptr0
                        sar     parm1, #2
                        call    #push
                        jmp     #innerloop

'*******************************************************************************
' Drop the top value from the stack
' No changes
'*******************************************************************************
dropfunc                sub     stackptr, #4
                        jmp     #innerloop

'*******************************************************************************
' Use the value on top of the stack as an index to another value in the stack,
' and write its value to the stack
' No changes
'*******************************************************************************
pickfunc                call    #pop
                        call    #indexstack
                        call    #push
                        jmp     #innerloop

'*******************************************************************************
' Use the value on top of the stack as and index to remove another value from
' the stack, and place it at the top of the stack.
' Changes temp1, temp2, temp3, temp4
'*******************************************************************************
rollfunc                call    #pop
                        cmp     parm1, #0               wc, wz
        if_c_or_z       jmp     #innerloop
                        mov     temp3, parm1
                        call    #indexstack
                        mov     temp2, temp1
:loop                   add     temp2, #4
                        rdlong  temp4, temp2
                        wrlong  temp4, temp1
                        add     temp1, #4
                        djnz    temp3, #:loop
                        wrlong  parm1, temp1
                        jmp     #innerloop

'*******************************************************************************
' Pop the value from the top of the stack, and push it onto the return stack.
' No changes
'*******************************************************************************
torfunc                 call    #pop
                        wrlong  parm1, returnptr
                        add     returnptr, #4
                        jmp     #innerloop

'*******************************************************************************
' Pop the value from the top of the return stack and push it to the stack.
' Changes parm1
'*******************************************************************************
fromrfunc               sub     returnptr, #4
                        rdlong  parm1, returnptr
                        call    #push
                        jmp     #innerloop

'*******************************************************************************
' Push the address of the input buffer onto the stack
' Changes parm1
'*******************************************************************************
tibfunc                 mov     parm1, a_inputbuf
                        call    #push
                        jmp     #innerloop

'*******************************************************************************
' Push the value on the stack pointed to by the PC and increment the PC
' Changes parm1
'*******************************************************************************
_litfunc                rdlong  parm1, pc
                        add     pc, #4
                        call    #push
                        jmp     #innerloop

'*******************************************************************************
' Convert the string described by the address and length on the top of the
' stack to a hex number, and push it to the stack
' Changes parm1
'*******************************************************************************
_gethexfunc             call    #pop2
                        call    #gethex
                        mov     parm1, parm
                        call    #push
                        jmp     #innerloop

'*******************************************************************************
' Create a variable, and add it to the dictionary
' Changes parm3
'*******************************************************************************
createfunc              mov     parm3, #varfunc
                        call    #create
                        jmp     #innerloop

'*******************************************************************************
' Create an executable word, and add it to the dictionary.  Set the compile
' state to -1
' Changes parm3, temp1
'*******************************************************************************
colonfunc               mov     parm3, #innerloopfunc
                        call    #create
        if_z            jmp     #innerloop
                        neg     temp1, #1
                        wrlong  temp1, a_state
                        jmp     #innerloop

'*******************************************************************************
' Compile a zero into memory indicating the end of an executable word, and set
' the compile flag to zero
' Changes temp1, temp2
'*******************************************************************************
semicolonfunc           mov     temp1, #0
                        wrlong  temp1, a_state
                        rdlong  temp2, a_dp
                        wrlong  temp1, temp2
                        add     temp2, #4
                        wrlong  temp2, a_dp
                        jmp     #innerloop

'*******************************************************************************
' Fetch a value from the specified cog address, and put it on the stack
' the compile flag to zero
' Changes parm1
'*******************************************************************************
cogfetchfunc            call    #pop
                        movs    cogfetch1, parm1
                        nop
cogfetch1               mov     parm1, 0-0
                        call    #push
                        jmp     #innerloop

'*******************************************************************************
' Get a cog address and value from the stack, and store the value at the address
' the compile flag to zero
' Changes parm1, parm2
'*******************************************************************************
cogstorefunc            call    #pop2
                        movd    cogstore1, parm2
                        nop
cogstore1               mov     0-0, parm1
                        jmp     #innerloop

'*******************************************************************************
' Print out an 8-digit hex number to the output port.
' Changes parm
'*******************************************************************************
dotxfunc                mov     parm, #"$"
                        call    #putch
                        call    #pop
                        call    #printhex
                        mov     parm, #" "
                        call    #putch
                        jmp     #innerloop

'*******************************************************************************
' The following code implements the basic functions used by the kernel words
'*******************************************************************************

'*******************************************************************************
' Create a word entry in the dictionary
' Changes parm, parm1, parm2, temp1, temp2              
'*******************************************************************************
create                  mov     parm, #" "
                        call    #word_del
        if_z            jmp     create_ret
                        rdlong  temp1, a_dp
                        add     temp1, #3
                        and     temp1, minus4
                        rdlong  temp2, a_last
                        wrlong  temp2, temp1
                        wrlong  temp1, a_last
                        add     temp1, #4
                        wrlong  parm3, temp1
                        add     temp1, #4
                        mov     temp2, #0
                        wrlong  temp2, temp1
                        add     temp1, #4
                        wrbyte  temp2, temp1
                        add     temp1, #1
                        wrbyte  parm2, temp1
                        add     temp1, #1
                        cmps    parm2, #0               wc, wz
        if_c_or_z       jmp     #create_done
:loop                   rdbyte  temp2, parm1
                        add     parm1, #1
                        wrbyte  temp2, temp1
                        add     temp1, #1               wz
                        djnz    parm2, #:loop
create_done             add     temp1, #3
                        and     temp1, minus4
                        wrlong  temp1, a_dp
create_ret              ret

'*******************************************************************************
' Write a long value at the data pointer, and increment the data pointer
' Assume that the data pointer is long-aligned
' Input parm
' Changes temp1
' Ouput none
'*******************************************************************************
compile                 rdlong  temp1, a_dp
                        wrlong  parm, temp1
                        add     temp1, #4
                        wrlong  temp1, a_dp
compile_ret             ret

'*******************************************************************************
' Refill the input buffer
' Input none
' Changes parm1, parm2
' getstr changes parm, parm1, parm2, temp, temp1, temp2, temp3, temp4
' Output none
'*******************************************************************************
refill                  mov     parm1, a_inputbuf
                        mov     parm2, #200
                        call    #getstr
                        wrlong  parm1, a_inputlen
                        mov     parm1, #0
                        wrlong  parm1, a_inputidx
refill_ret              ret

'*******************************************************************************
' Get a string from the input port.  Terminate after receiving a LF, CR or
' the specified number of characters.
' Input parm1, parm2
' Changes parm1, parm2, temp3, temp4
' getch changes parm, temp, temp1, temp2
' putch changes parm, temp1, temp2
' Output parm1
'*******************************************************************************
getstr                  mov     temp3, parm1
                        mov     parm1, #0
                        cmps    parm2, #0               wc, wz
        if_c_or_z       jmp     getstr_ret
:loop                   call    #getch
                        mov     temp4, parm
                        call    #putch
                        cmp     temp4, #13              wz
        if_z            jmp     getstr_ret
                        cmp     temp4, #10              wz
        if_z            jmp     getstr_ret
                        wrbyte  temp4, temp3
                        add     parm1, #1
                        add     temp3, #1
                        djnz    parm2, #:loop
getstr_ret              ret

'*******************************************************************************
' Write a string to the output port.
' Input parm1, parm2
' Changes parm, parm1, parm2
' putch changes parm, temp1, temp2
' Output none
'*******************************************************************************
putstr                  cmps    parm2, #0               wc, wz
        if_c_or_z       jmp     putstr_ret
:loop                   rdbyte  parm, parm1
                        call    #putch
                        add     parm1, #1
                        djnz    parm2, #:loop
putstr_ret              ret              

'*******************************************************************************
' Get one character from the input port.
' Input none
' Changes parm, temp, temp1, temp2
' Output parm
'*******************************************************************************
getch                   rdbyte  parm, infileptr         wz
        if_z            jmp     #getch1
                        add     infileptr, #1
                        jmp     getch_ret
getch1                  mov     parm, ina
                        and     parm, inbit             wz
        if_nz           jmp     #getch
                        mov     temp2, cnt
                        mov     temp, bitcycles
                        shr     temp, #1
                        add     temp2, temp
                        mov     temp1, #10
:loop                   waitcnt temp2, bitcycles
                        mov     temp, ina
                        and     temp, inbit
                        ror     parm, #1
                        or      parm, temp
                        djnz    temp1, #:loop
                        ror     parm, #31 - 8
                        and     parm, #255
getch_ret               ret

inbit                   long    $80000000
bitcycles               long    80_000_000 / 115_200

'*******************************************************************************
' Send one character to the output port.
' Input parm
' Changes parm, temp1, temp2
' Output none             
'*******************************************************************************
putch                   or      parm, #$100
                        shl     parm, #1
                        rol     parm, #30
                        mov     temp1, #10
                        mov     temp2, bitcycles
                        add     temp2, cnt
:loop                   mov     outa, parm
                        ror     parm, #1
                        waitcnt temp2, bitcycles
                        djnz    temp1, #:loop
putch_ret               ret

outbit                  long    $40000000

'*******************************************************************************
' Skip the specified character in the input buffer
' Input parm
' Changes temp, temp1
' Output none
'*******************************************************************************
skipchar                cmps    temp1, temp2            wc
        if_nc           jmp     skipchar_ret
                        mov     temp, a_inputbuf
                        add     temp, temp1
                        rdbyte  temp, temp
                        cmp     temp, parm              wz
        if_nz           jmp     skipchar_ret
                        add     temp1, #1
                        jmp     #skipchar
skipchar_ret            ret

'*******************************************************************************
' Find the next occurance of the specified character in the input buffer
' Input parm
' Changes temp, temp1
' Output none
'*******************************************************************************
findchar                cmps    temp1, temp2            wc
        if_nc           jmp     findchar_ret
                        mov     temp, a_inputbuf
                        add     temp, temp1
                        rdbyte  temp, temp
                        cmp     temp, parm              wz
        if_z            jmp     findchar_ret
                        add     temp1, #1
                        jmp     #findchar
findchar_ret            ret

'*******************************************************************************
' Find the next word in the input buffer delimited by the specified character
' Input parm
' Changes parm1, parm2, temp1, temp2
' Output none
'*******************************************************************************
word_del
                        rdlong  temp1, a_inputidx
                        rdlong  temp2, a_inputlen
                        call    #skipchar
                        mov     parm1, temp1
                        call    #findchar
                        mov     parm2, temp1
                        sub     parm2, parm1            wz
                        add     parm1, a_inputbuf
                        cmps    temp1, temp2            wc
        if_c            add     temp1, #1
                        wrlong  temp1, a_inputidx
word_del_ret            ret

'*******************************************************************************
' Find the specified word in the dictionary
' Input parm1, parm2
' Changes parm, parm3, parm4
' Output parm
'*******************************************************************************
findword                rdlong  parm, a_last            wz
        if_z            jmp     findword_ret
:loop                   mov     parm3, parm
                        add     parm3, #13
                        rdbyte  parm4, parm3
                        add     parm3, #1
                        call    #compare
        if_z            jmp     findword_ret
                        rdlong  parm, parm              wz
        if_nz           jmp     #:loop
findword_ret            ret

'*******************************************************************************
' Do a case insensitive comparison of two character strings
' Input parm1, parm2, parm3, parm4
' Changes parm3, parm4, temp, temp1, temp2
' Outut Z
'*******************************************************************************
compare                 cmps    parm2, #1               wc, wz
        if_c            jmp     compare_ret
                        cmp     parm2, parm4            wz
        if_nz           jmp     compare_ret
                        mov     temp, parm1
:loop                   rdbyte  temp1, temp
                        call    #toupper
                        mov     temp2, temp1
                        rdbyte  temp1, parm3
                        call    #toupper
                        cmp     temp1, temp2            wz
        if_nz           jmp     compare_ret
                        add     temp, #1
                        add     parm3, #1
                        djnz    parm4, #:loop
compare_ret             ret

'*******************************************************************************
' Convert a character to uppercase
' Input temp1
' Changes temp1
' Ouput temp1
'*******************************************************************************
toupper                 cmp     temp1, #"a"             wc
        if_c            jmp     toupper_ret
                        cmp     temp1, #"z"             wc, wz
        if_nc_and_nz    jmp     toupper_ret
                        sub     temp1, #"a" - "A"
toupper_ret             ret              

'*******************************************************************************
' Print an 8-digit hex value to the output port
' Input parm1
' Changes parm, parm1, parm2
' Output none
'*******************************************************************************
printhex                mov     parm2, #8
:loop                   rol     parm1, #4
                        mov     parm, #15
                        and     parm, parm1
                        add     parm, a_hexstr
                        rdbyte  parm, parm
                        call    #putch
                        djnz    parm2, #:loop
printhex_ret            ret

'*******************************************************************************
' Convert a string to a hex number
' Input parm1, parm2
' Changes parm, temp, temp1, temp2
' Output parm
'*******************************************************************************
gethex                  mov     parm, #0
                        cmps    parm2, #0               wc, wz
        if_c_or_z       jmp     gethex_ret
                        mov     temp1, parm1
                        mov     temp2, parm2
:loop                   rdbyte  temp, temp1
                        add     temp1, #1
                        sub     temp, #"0"
                        cmps    temp, #10               wc
        if_nc           sub     temp, #"a"-"0"-10
                        shl     parm, #4
                        add     parm, temp
                        djnz    temp2, #:loop
gethex_ret              ret

'*******************************************************************************
' Push a value onto the data stack
' Input parm1
' No changes
' Output none
'*******************************************************************************
push                    wrlong  parm1, stackptr
                        add     stackptr, #4
push_ret                ret

'*******************************************************************************
' Pop two values off of the data stack
' Input none
' Changes parm1, parm2
' Output parm1, parm2
'*******************************************************************************
pop2                    sub     stackptr, #4
                        rdlong  parm2, stackptr

'*******************************************************************************
' Pop one value off of the data stack
' Input none
' Changes parm1
' Ouput parm1
'*******************************************************************************
pop                     sub     stackptr, #4
                        rdlong  parm1, stackptr
pop_ret
pop2_ret                ret

'*******************************************************************************
' Read a value on the stack based on an index number
' Changes parm1, temp1
'*******************************************************************************
indexstack              neg     temp1, parm1
                        shl     temp1, #2
                        sub     temp1, #4
                        add     temp1, stackptr
                        rdlong  parm1, temp1
indexstack_ret          ret

'*******************************************************************************
' Compute the address of the body of a word base on it's starting address plus 4
' Changes parm, parm1
'*******************************************************************************
body1                   add     parm, #9
                        rdbyte  parm1, parm             ' Get name length
                        add     parm1, parm
                        add     parm1, #4
                        and     parm1, minus4           ' Align
body1_ret               ret

'*******************************************************************************
' Multiply two 32-bit numbers
' Changes parm2, temp1, temp2
'*******************************************************************************
multiply                mov     temp1, #0
                        mov     temp2, #32
                        shr     parm1, #1               wc
mmul    if_c            add     temp1, parm2            wc
                        rcr     temp1, #1               wc
                        rcr     parm1, #1               wc
                        djnz    temp2, #mmul
multiply_ret            ret

'*******************************************************************************
' Divide two 32-bit numbers producing a quotient and a remainder
' Changes parm1, parm2, parm3, temp1, temp2
'*******************************************************************************
divide                  mov     temp2, #32
                        mov     temp1, #0
                        abs     parm1, parm1            wc
                        muxc    parm3, #%01100
                        abs     parm2, parm2            wc,wz
        if_c            xor     parm3, #%00100
        if_nz           jmp     #mdiv
                        mov     parm1, #0
                        jmp     divide_ret
mdiv                    shr     parm2, #1               wc,wz
                        rcr     temp1, #1
        if_nz           djnz    temp2, #mdiv
mdiv2                   cmpsub  parm1, temp1            wc
                        rcl     parm2, #1
                        shr     temp1, #1
                        djnz    temp2, #mdiv2
divide_ret              ret

'*******************************************************************************
' These are working registers.  The parm registers are generally used to pass
' parameters from one routine to another, and the temp registers are used as
' temporary storage within a routine.
'*******************************************************************************
parm1                   long      0
parm2                   long      0
parm3                   long      0
parm4                   long      0
temp1                   long      0
temp2                   long      0
temp3                   long      0
temp4                   long      0

'*******************************************************************************
' Addresses of variables in the dictionary, and the hex table
'*******************************************************************************
a_hexstr                long      @hexstr+16
a_last                  long      @last+16
a_state                 long      @state+16
a_dp                    long      @dp+16
a_inputbuf              long      @inputbuf+16
a_inputidx              long      @greaterin+16
a_inputlen              long      @poundtib+16

'*******************************************************************************
' The data and return stack pointers, and their base addresses
'*******************************************************************************
stackptr                long      @stack+16+16
stackptr0               long      @stack+16+16
returnptr               long      @retstk+16
returnptr0              long      @retstk+16

'*******************************************************************************
' The input file pointer used during initialization
'*******************************************************************************
infileptr               long      @infile+16

'*******************************************************************************
' Constants
'*******************************************************************************
minus4                  long      -4

                        fit       $1f0

'*******************************************************************************
' Data stack, return stack, input buffer and hex table
'*******************************************************************************
hexstr                  byte      "0123456789abcdef"
inputbuf                byte      0[200]
stack                   long      0[200]
retstk                  long      0[200]

'*******************************************************************************
' This is the beginning of the dictionary.  The kernel words are specified below
'*******************************************************************************
_exit         long      0, exitfunc, 0
              byte      0, 4, "exit"
_quit         long      @_exit+16, quitfunc, 0
              byte      0, 4, "quit"
_abort        long      @_quit+16, abortfunc, 0
              byte      0, 5, "abort"
_execute      long      @_abort+16, executefunc, 0
              byte      0, 7, "execute"
_refill       long      @_execute+16, refillfunc, 0
              byte      0, 6, "refill"
_word         long      @_refill+16, wordfunc, 0
              byte      0, 4, "word"
_find         long      @_word+16, findfunc, 0
              byte      0, 4, "find"
_key          long      @_find+16, keyfunc, 0
              byte      0, 3, "key"
_accept       long      @_key+16, acceptfunc, 0
              byte      0, 6, "accept"
_create       long      @_accept+16, createfunc, 0
              byte      0, 6, "create"
__lit         long      @_create+16, _litfunc, 0
              byte      0, 4, "_lit"
__gethex      long      @__lit+16, _gethexfunc, 0
              byte      0, 7, "_gethex"
_emit         long      @__gethex+16, emitfunc, 0
              byte      0, 4, "emit"
_store        long      @_emit+16, storefunc, 0
              byte      0, 1, "!"
_fetch        long      @_store+16, fetchfunc, 0
              byte      0, 1, "@"
_cstore       long      @_fetch+16, cstorefunc, 0
              byte      0, 2, "c!"
_cfetch       long      @_cstore+16, cfetchfunc, 0
              byte      0, 2, "c@"
_plus         long      @_cfetch+16, plusfunc, 0
              byte      0, 1, "+"
_minus        long      @_plus+16, minusfunc, 0
              byte      0, 1, "-"
_multiply     long      @_minus+16, multfunc, 0
              byte      0, 1, "*"
_divide       long      @_multiply+16, dividefunc, 0
              byte      0, 1, "/"
_mod          long      @_divide+16, modfunc, 0
              byte      0, 3, "mod"
_and          long      @_mod+16, andfunc, 0
              byte      0, 3, "and"
_or           long      @_and+16, orfunc, 0
              byte      0, 2, "or"
_xor          long      @_or+16, xorfunc, 0
              byte      0, 3, "xor"
_less         long      @_xor+16, lessfunc, 0
              byte      0, 1, "<"
_equal        long      @_less+16, equalfunc, 0
              byte      0, 1, "="
_greater      long      @_equal+16, greaterfunc, 0
              byte      0, 1, ">"
_rshift       long      @_greater+16, rshiftfunc, 0
              byte      0, 6, "rshift"
_lshift       long      @_rshift+16, lshiftfunc, 0
              byte      0, 6, "lshift"
_depth        long      @_lshift+16, depthfunc, 0
              byte      0, 5, "depth"
_tib          long      @_depth+16, tibfunc, 0
              byte      0, 3, "tib"
_poundtib     long      @_tib+16, varfunc, 0
              byte      0, 4, "#tib"
poundtib      long      0
_greaterin    long      @_poundtib+16, varfunc, 0
              byte      0, 3, ">in"
greaterin     long      0
_dp           long      @_greaterin+16, varfunc, 0
              byte      0, 2, "dp"
dp            long      @___here+16
_last         long      @_dp+16, varfunc, 0
              byte      0, 4, "last"
last          long      @___last+16
_state        long      @_last+16, varfunc, 0
              byte      0, 5, "state"
state         long      0
_base         long      @_state+16, varfunc, 0
              byte      0, 4, "base"
base          long      16
_drop         long      @_base+16, dropfunc, 0
              byte      0, 4, "drop"
_pick         long      @_drop+16, pickfunc, 0
              byte      0, 4, "pick"
_roll         long      @_pick+16, rollfunc, 0
              byte      0, 4, "roll"
_tor          long      @_roll+16, torfunc, 0
              byte      0, 2, ">r"
_fromr        long      @_tor+16, fromrfunc, 0
              byte      0, 2, "r>"
_colon        long      @_fromr+16, colonfunc, 0
              byte      0, 1, ":"
_semicolon    long      @_colon+16, semicolonfunc, 0
              byte      1, 1, ";"
_cogfetch     long      @_semicolon+16, cogfetchfunc, 0
              byte      0, 4, "cog@"
_cogstore     long      @_cogfetch+16, cogstorefunc, 0
              byte      0, 4, "cog!"
_dotx         long      @_cogstore+16, dotxfunc, 0
              byte      0, 2, ".x"
              
'*******************************************************************************
' A small number of compiled words follow below.  These are used by the boot
' interpreter.
'*******************************************************************************
              ' : here dp @ ;
_here         long      @_dotx+16, innerloopfunc, 0
              byte      0, 4, "here"
              long      @_dp+16, @_fetch+16, 0

              ' : allot dp @ + dp ! ;
_allot        long      @_here+16, innerloopfunc, 0
              byte      0, 5, "allot"
              long      @_dp+16, @_fetch+16, @_plus+16, @_dp+16, @_store+16, 0

              ' : , here ! 4 allot ;
_comma        long      @_allot+16, innerloopfunc, 0
              byte      0, 1, ","
              long      @_here+16, @_store+16, @__lit+16, 4, @_allot+16, 0

              ' : _jmp r> @ >r ;
__jmp         long      @_comma+16, innerloopfunc, 0
              byte      $A, 4, "_jmp"
              long      @_fromr+16, @_fetch+16, @_tor+16, 0

              ' : _jz r> 0 pick 4 + 1 roll @ 2 roll 0 = 1 + pick >r drop drop ;
__jz          long      @__jmp+16, innerloopfunc, 0
              byte      $A, 3, "_jz"
              long      @_fromr+16, @__lit+16, 0, @_pick+16, @__lit+16, 4, @_plus+16, @__lit+16, 1, @_roll+16, @_fetch+16, @__lit+16, 2, @_roll+16
              long      @__lit+16, 0, @_equal+16, @__lit+16, 1, @_plus+16, @_pick+16, @_tor+16, @_drop+16, @_drop+16, 0

              ' : count 0 pick 1 + 1 roll c@ ;
_count        long      @__jz+16, innerloopfunc, 0
              byte      0, 5, "count"
              long      @__lit+16, 0, @_pick+16, @__lit+16, 1, @_plus+16, @__lit+16, 1, @_roll+16, @_cfetch+16, 0

___last       long

'*******************************************************************************
' The boot interpreter follows below.
'*******************************************************************************
              ' : xboot ( This word runs a simple interpreter )
_xboot        long      @_count+16, innerloopfunc, 0
              byte      0, 5, "xboot"

              ' 20 word 0 pick c@ _jz _xboot2 ( Get word, refill if empty )
_xboot1       long      @__lit+16, $20, @_word+16, @__lit+16, 0, @_pick+16, @_cfetch+16, @__jz+16, @_xboot2+16
              
              ' find 0 pick _jz _xboot3 ( Find word, get number if not found )
              long      @_find+16, @__lit+16, 0, @_pick+16, @__jz+16, @_xboot3+16
              
              ' state @ = _jz _xboot4 ( Go execute if not compile mode or immediate )
              long      @_state+16, @_fetch+16, @_equal+16, @__jz+16, @_xboot4+16
              
              ' , _jmp _xboot1 ( Otherwise, compile and loop again )
              long       @_comma+16, @__jmp+16, @_xboot1+16
              
              ' execute _jmp _xboot1 ( Execute and loop again )
_xboot4       long      @_execute+16, @__jmp+16, @_xboot1+16

              ' drop count _gethex ( Get number )
_xboot3       long      @_drop+16, @_count+16, @__gethex+16

              ' state @ _jz _xboot1 ( Loop again if not compile mode )
              long      @_state+16, @_fetch+16, @__jz+16, @_xboot1+16
              
              ' ['] _lit , , _jmp _xboot1 ( Otherwise, compile number and loop again )
              long       @__lit+16, @__lit+16, @_comma+16, @_comma+16, @__jmp+16, @_xboot1+16

              ' drop refill _jmp _xboot1 ( Refill and loop again )
_xboot2       long      @_drop+16, @_refill+16, @__lit+16, 13, @_emit+16, @__jmp+16, @_xboot1+16, 0

___here       long

'*******************************************************************************
' The Forth source files follow below.  They will be compiled into the
' dictionary, which will over-write the source data.  Some padding space is
' included to ensure that we don't over-write the source data before it is
' compiled.  The source data is terminated by a zero to indicate the
' end of the data.
'*******************************************************************************
              long      0[1000]
infile        file      "init.fth"
              file      "chkcore.fth"
              file      "chkcorex.fth"
              file      "misc.fth"
              file      "starting.fth"
              file      "rc4time.fth"
              long      0

{
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
}